/* interp.c */
/* this file contains interpreter routines which can be called
   by the keyboard interpreter (ki) or the real-time scheduler */

#include <dos.h>
#include <stdio.h>
#include <at200.h>

#include "idecls.h"
#include "proto.h"
#include "camsys.h"
#include "mode.h"

/* Globals used in this file. */

static int  bufp;
static char nbuf[NBUFSIZE]; /* Interpreter name table */

/* Routine to delete previous additions made to the symbol table*/

void delete( void )
{
	register int i ;
	bufp = 0 ;
	for ( i=0 ; i <= STSIZE ; ++i )
	switch ( symtab[i].type )
	{
		case UPROC:
		case VARIABLE:
		case CONST:
		case LITERAL:
			symtab[i].type = UNDEF ;
	}
}

/* Interpreter parsing module. Passed a pointer s to input character
   string. Separates line into tokens and produces output array of 
   program vector pointers in prv[]. Sets integer pointed to by 'next'
   to next available index into prv[] (i.e. returns 0 if no tokens
   found). Unrecognized alphanumeric tokens are considered integer
   variables unless preceded by # in which case they are assumed
   to define a user procedure name. In this case a NULL pointer is
   entered in prv[] and the name is entered in the symbol table
    with type UPROC and a value consisting of a pointer to the
   next entry in prv[]. Note that the keyboard interpreter does
   not support user procedure definitions. */

char *iparse ( char *s, entry_ptr prv[], register int *next )
{
	register int i, j;
	int flag;
	char token[ MAXTOKEN ];

	for ( i=0 ; &prv[i] < &prv[PVSIZE] ; i++ )
	{
		s = get_token ( s, token, &flag ) ;
		if ( flag == NOT_NUMERIC_ERROR ) 
			return ("Numeric value expected.\n") ;
		if ( flag == NOT_HEX_ERROR ) 
			return ("Hexadecimal value expected.\n") ;
		if ( flag == NOT_TERMINATED_ERROR ) 
			return ("Closing quote expected.\n") ;
		if ( flag == TOKEN_LEN_ERROR ) 
			return ("Token too long.\n") ;
		if ( flag == NOTFOUND )
		{
			*next = i ;
			return ( NULL ) ;
		}
		if (( j = isearch ( token, flag )) >= STSIZE )
			return ("Symbol table full.\n") ;
		if ( symtab[j].type == UNDEF )
		{
			if ((symtab[j].name = nsave(token)) == NULL )
				return ("Name buffer full.\n") ;
			switch (flag)
			{
			case ALPHA:
				symtab[j].type = VARIABLE ;
				symtab[j].uval.value = 0 ;
				break ;
			case NUMERIC:
				symtab[j].type = CONST ;
				symtab[j].uval.value = convert(token) ;
				break ;
			case LITERAL:
				symtab[j].type = LITERAL ;
				symtab[j].uval.value = 0 ;
				break ;
			case UPROC:
				symtab[j].type = UPROC ;
				symtab[j].uval.uproc = &prv[i+1] ;
				break ;
			}
		}
		prv[i] = (flag == UPROC) ? NULL : &symtab[j] ;
	}
	return ("Program vector overflow.\n") ;
}

/* Save string s in nbuf. Returns pointer to first char of saved
   string or NULL if buffer full. */

char *nsave ( register char *s ) 
{
	register char *rval ;
	rval = &nbuf[bufp] ;
	while ((nbuf[bufp++] = *s++) != '\0')
		if (bufp >= NBUFSIZE) return (NULL) ;
	return (rval) ;
}

/* Search for a token in the symbol table. Returns index of entry
   in symtab if found, or next free symtab location if not. */

int isearch ( register char *ts, register int typ )
{
	register int i ;
	char *jp, *kp ;
	for (i=0 ; i <= STSIZE ; i++ )
	{
		if (symtab[i].type == UNDEF ) return (i) ;
		for (jp=ts, kp=symtab[i].name; *jp == *kp ; jp++, kp++ )
			if (*jp == '\0')
			{
				if (typ == LITERAL)
				{ if (symtab[i].type==LITERAL) return(i) ;}
				else if (symtab[i].type != LITERAL) return (i);
			}
	}
	return (i) ;
}
				

/* routine to get next token starting at sp and put it in ts,
   along with \0. Sets flag to ALPHA, NUMERIC, LITERAL, NOTFOUND
   or ERROR. Returns a pointer to next char in sp following token. */

char *get_token ( register char *sp, register char *ts, int *flag )
{
	register int c ;
	int	token_len = 0;
	*flag = NOTFOUND ;
	while ( ( c = *sp++ ) != '\n' && c != EOF && c != '\0' )
		switch ( chtest ( c ) )
		{
		case ALPHA :
			*ts++ = (char) c;
			if ( ++token_len == MAXTOKEN )
			{
				*flag = TOKEN_LEN_ERROR;
				return (sp);
			}

			while ( ( c = chtest (*sp)) == NUMERIC || c == ALPHA)
			{
				*ts++ = *sp++;
				if ( ++token_len == MAXTOKEN )
				{
					*flag = TOKEN_LEN_ERROR;
					return (sp);
				}
			}
			*flag = ALPHA ;
			*ts = '\0' ;
			return (sp) ;
		case NUMERIC :
			*ts++ = (char) c;
			if ( ++token_len == MAXTOKEN )
			{
				*flag = TOKEN_LEN_ERROR;
				return (sp);
			}

			while ( chtest(*sp) == NUMERIC )
			{
				*ts++ = *sp++;
				if ( ++token_len == MAXTOKEN )
				{
					*flag = TOKEN_LEN_ERROR;
					return (sp);
				}
			}
			*flag = NUMERIC ;
			*ts = '\0' ;
			return (sp) ;
		case '-' :
			*ts++ = (char) c ;
			if ( ++token_len == MAXTOKEN )
			{
				*flag = TOKEN_LEN_ERROR;
				return (sp);
			}

			if ( chtest(*sp) != NUMERIC)
			{
				*flag = NOT_NUMERIC_ERROR ;
				return (sp) ;
			}
			do 
			{
				*ts++ = *sp++;
				if ( ++token_len == MAXTOKEN )
				{
					*flag = TOKEN_LEN_ERROR;
					return (sp);
				}
			} while (chtest(*sp) == NUMERIC) ;
			*flag = NUMERIC ;
			*ts = '\0' ;
			return (sp) ;
		case '%' :
			*ts++ = (char) c ;
			if ( ++token_len == MAXTOKEN )
			{
				*flag = TOKEN_LEN_ERROR;
				return (sp);
			}

			if ( !ishex(*sp))
			{
				*flag = NOT_HEX_ERROR ;
				return (sp) ;
			}
			do
			{
				*ts++ = *sp++ ;
				if ( ++token_len == MAXTOKEN )
				{
					*flag = TOKEN_LEN_ERROR;
					return (sp);
				}
			} while (ishex(*sp)) ;
			*flag = NUMERIC ;
			*ts = '\0' ;
			return (sp) ;
		case '"':
		case '\'' :
			while (( c = *sp++ ) != '\'' && c != '"')
			{
				if(( c == '\n') || (c == EOF) || (c == '\0' ))
				{
					*flag = NOT_TERMINATED_ERROR ;
					return (sp) ;
				}
				*ts++ = (char) c ;
				if ( ++token_len == MAXTOKEN )
				{
					*flag = TOKEN_LEN_ERROR;
					return (sp);
				}
			}
			*flag = LITERAL ;
			*ts = '\0' ;
			return (sp) ;
		case '{' :
			*ts++ = 'b'; 
			*ts++ = 'e'; 
			*ts++ = 'g' ; 
			*ts++ = 'i' ; 
			*ts++ = 'n';
			*ts   = '\0' ;
			*flag = ALPHA ;
			return (sp) ;
		case '}' :
			*ts++ = 'e' ; 
			*ts++ = 'n' ; 
			*ts++ = 'd' ; 
			*ts   = '\0' ;
			*flag = ALPHA ;
			return (sp) ;
		}
}

/* test a character and return ALPHA, NUMERIC or the character
   itself if neither. */

int chtest ( register int c ) 
{
	if ((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z'))
	return (ALPHA) ;
	return ( c >= '0' && c <= '9' ? NUMERIC : c ) ;
}

/* test a character to see whether hex. Returns 1 if it is, else 0 */

int ishex ( int c )
{
	return ((c>= '0' && c <= '9') || (c >= 'a' && c <= 'f')
			|| (c >= 'A' && c <= 'F') ? 1 : 0 ) ;
}

/* Convert an ASCII string to integer. Input parameter is
   a pointer to the string, returns the converted value. 
   Number is negative if preceded by minus sign, or hex
   if preceded by '%' */

int convert( char *s )
{
	register int i, n, sign ;
	sign = 1 ;
	i = 0 ;
	if ( s[0] == '%' )
		for (n=0,i=1; s[i] != '\0'; i++)
		n = 16*n + htoi(s[i]) ;
	else
	{
		if (s[0] == '-')
		{
			sign = -1 ;
			i = 1 ;
		}
		for (n=0; s[i] != '\0'; i++)
		n = 10 * n + s[i] - '0' ;
	}
	return (sign * n) ;
}

/* convert hex character c to integer */

int htoi ( register int c )
{
	if ( c >= '0' && c <= '9' ) return (c - '0') ;
	if ( c >= 'a' && c <= 'f' ) return (c - 'a' + 10) ;
	return ( c - 'A' + 10 ) ;
}



/* Execute a program vector by calling "clause" until NULL 
   encountered in program vector, or until global variable
   "abrt" becomes other than NORMAL. Takes a single input
   parameter of type "pv_ptr" which points to the element
   in the program vector where execution is to start. */

void iexec ( register pv_ptr p )
{
	int t, v ;

	while ( abrt == NORMAL )
	{
		if ( *p == NULL )    return;
		p = clause ( p, &t, &v );
	}
}

/* Core routine for executing program vectors. Clause is passed
   a pointer to an element in a program vector. The element is
   examined, and if it is a language procedure ( i.e. defined
   in isym.c), the procedure is called. If the element is a
   user defined procedure (i.e. using the # procedure definition
   capability) then that function is executed by calling "clause".
   Otherwise "clause" returns the value of the element. */

pv_ptr clause ( register pv_ptr p, int *t, int *v )
{
	pv_ptr (*fptr)();
	if (*p == NULL || abrt != NORMAL) return (p) ;
	switch ((**p).type)
	{
	case LPROC:
		fptr = (**p).uval.lproc ;
		return ((*fptr)(++p)) ;
	case UPROC:
		return ( clause ( (**p).uval.uproc, t, v ) ) ;
	default:
		{
			*t = (**p).type ;
			*v = (**p).uval.value ;
			return (++p) ;
		}
	}
}
 

